/* GNOME DB
 * Copyright (C) 1998 Michael Lausch
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public
 * License along with this library; if not, write to the Free
 * Software Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */



#include <memory.h>
#include <glib.h>

#include "odbc.h"
#include "sqldriver.h"

#include "gda-types.h"
#include "gda-connection.h"
#include "gda-command.h"
#include "gda-error.h"
#include "gda-command.h"

typedef Gda_ODBC_Recordset* (*schema_ops_fn)(Gda_ODBC_Recordset*,
					     Gda_ODBC_Error*,
					     Gda_ODBC_Connection* ,
					     GDA_Connection_Constraint* constraints,
					     gint length);

schema_ops_fn schema_ops[GDNC_SCHEMA_LAST] =
{
  0,
};

static void            init_gda_odbc_lib                (void) __attribute__((constructor));
static void            initialize_schema_ops            (void);
static Gda_ODBC_Recordset*  schema_tables(Gda_ODBC_Recordset*  recset,
					  Gda_ODBC_Error*      e,
					  Gda_ODBC_Connection* cnc,
					  GDA_Connection_Constraint* constraints,
					  gint length);

static Gda_ODBC_Recordset*  schema_columns(Gda_ODBC_Recordset*  recset,
					   Gda_ODBC_Error*      e,
					   Gda_ODBC_Connection* cnc,
					   GDA_Connection_Constraint* constraints,
					   gint length);
static Gda_ODBC_Recordset*  schema_types(Gda_ODBC_Recordset*   recset,
					 Gda_ODBC_Error*       e,
					 Gda_ODBC_Connection*  cnc,
					 GDA_Connection_Constraint*  constraint,
					 gint length);


static void
initialize_schema_ops(void)
{
  schema_ops[GDCN_SCHEMA_TABLES] = schema_tables;
  schema_ops[GDCN_SCHEMA_COLS]   = schema_columns;
  schema_ops[GDCN_SCHEMA_PROV_TYPES] = schema_types;
}


static void
init_gda_odbc_lib(void)
{
  fprintf(stderr,"init_gda_odbc_lib called\n");
  initialize_schema_ops();
}


static Gda_ODBC_Recordset*
schema_tables(Gda_ODBC_Recordset*   recset,
	      Gda_ODBC_Error*       e,
	      Gda_ODBC_Connection*  cnc,
	      GDA_Connection_Constraint* constraints,
	      gint length)
{
  gint                rc;
  Gda_ODBC_Command*   cmd;
  
  cmd = gda_odbc_cmd_new();
  cmd->cnc = cnc;
  recset->cmd = cmd;
  rc = SQLAllocStmt(cnc->hdbc, &recset->hstmt);
  if (rc == SQL_ERROR)
    return 0;
  rc = SQLTables(recset->hstmt,
		 "", SQL_NTS,
		 "", SQL_NTS,
		 "", SQL_NTS,
		 "", SQL_NTS);
  if (rc == SQL_ERROR)
    {
      gda_odbc_error_make(gda_odbc_error_new(),
			  recset,
			  cnc,
			  "SQLTables");
      SQLFreeStmt(recset->hstmt, SQL_DROP);
      return 0;
    }
  return gda_odbc_init_recset_for_hstmt(recset, e);
}


static Gda_ODBC_Recordset*
schema_columns(Gda_ODBC_Recordset*  recset,
	       Gda_ODBC_Error*      e,
	       Gda_ODBC_Connection* cnc,
	       GDA_Connection_Constraint* constraints,
	       gint            length)
{
  gint             rc;
  Gda_ODBC_Command*     cmd;
  gchar*           table_qualifier = 0;
  gchar*           table_owner     = 0;
  gchar*           table_name      = 0;
  gchar*           column_name     = 0;
  GDA_Connection_Constraint*  ptr;
  
  cmd = gda_odbc_cmd_new();
  cmd->cnc = cnc;
  recset->cmd = cmd;
  rc = SQLAllocStmt(cnc->hdbc, &recset->hstmt);
  if (rc == SQL_ERROR)
    return 0;
  ptr = constraints;
  while(length)
    {
      switch (ptr->ctype)
	{
	case GDA_Connection_TABLE_CATALOG:
	  table_qualifier = ptr->value;
	  fprintf(stderr,"schema_columns: table_qualifier = '%s'\n", table_qualifier);
	  break;
	case GDA_Connection_TABLE_SCHEMA:
	  table_owner = ptr->value;
	  fprintf(stderr,"schema_columns: table_owner = '%s'\n", table_owner);
	  break;
	case GDA_Connection_TABLE_NAME:
	  table_name = ptr->value;
	  fprintf(stderr,"schema_columns: table_name = '%s'\n", table_name);
	  break;
	case GDA_Connection_COLUMN_NAME:
	  column_name = ptr->value;
	  fprintf(stderr,"schema_columns: column_name = '%s'\n", column_name);
	  break;
	default:
	  fprintf(stderr,"schema_columns: invalid constraint type %d\n", ptr->ctype);
	  SQLFreeStmt(recset->hstmt, SQL_DROP);
	  return 0;
	  break;
	}
      length--;
      ptr++;
    }
  rc = SQLColumns(recset->hstmt,
		  table_qualifier, table_qualifier ? SQL_NTS : 0,
		  table_owner,     table_owner ? SQL_NTS : 0, 
		  table_name,      table_name ? SQL_NTS : 0,
		  column_name,     column_name ? SQL_NTS : 0);
  if (rc == SQL_ERROR)
    {
      gda_odbc_error_make(gda_odbc_error_new(), recset, cnc, "SQLColumns");
      SQLFreeStmt(recset->hstmt, SQL_DROP);
      return 0;
    }
  fprintf(stderr,"schema_columns: calling gda_init_recset_for_hstmt\n");
  return gda_odbc_init_recset_for_hstmt(recset, e);
}

static Gda_ODBC_Recordset*
schema_types(Gda_ODBC_Recordset*   recset,
	     Gda_ODBC_Error*       e,
	     Gda_ODBC_Connection*  cnc,
	     GDA_Connection_Constraint*  constraints,
	     gint             length)
{
  gint              rc;
  Gda_ODBC_Command* cmd;

  cmd = gda_odbc_cmd_new();
  cmd->cnc = cnc;
  recset->cmd = cmd;
  rc = SQLAllocStmt(cnc->hdbc, &recset->hstmt);
  if (rc == SQL_ERROR)
    return 0;
  rc = SQLGetTypeInfo(recset->hstmt, 0);
  if (rc == SQL_ERROR)
    {
      gda_odbc_error_make(gda_odbc_error_new(), recset, cnc, "SQLColumns");
      SQLFreeStmt(recset->hstmt, SQL_DROP);
      return 0;
    }
  return gda_odbc_init_recset_for_hstmt(recset, e);
}




Gda_ODBC_Connection*
gda_odbc_connection_new()
{

  Gda_ODBC_Connection* c = g_new(Gda_ODBC_Connection, 1);
  memset(c, 0, sizeof(Gda_ODBC_Connection));
  return c;
}


void
gda_odbc_connection_free(Gda_ODBC_Connection* c)
{
  g_free(c);
}






gint
gda_odbc_connection_open(Gda_ODBC_Connection* cnc,
			 const gchar*               cnc_string,
			 const gchar*               user,
			 const gchar*               passwd)
{
  int rc;
  
  rc = SQLAllocEnv(&cnc->henv);
  if (rc != SQL_SUCCESS)
    return -1;

  rc = SQLAllocConnect(cnc->henv, &cnc->hdbc);
  if (rc != SQL_SUCCESS)
    {
      SQLFreeEnv(cnc->henv);
      return -1;
    }
  g_warning("gda_odbc_connection_open: cnc_string = '%s'\n", cnc_string);
  g_warning("gda_odbc_connection_open: user = '%s'\n", user);
  g_warning("gda_odbc_connection_open: passwd = '%s'\n", passwd);
  rc = SQLConnect(cnc->hdbc, (char*)cnc_string, SQL_NTS, (char*) user, SQL_NTS,
		  (char*)passwd, SQL_NTS);
  if (rc != SQL_SUCCESS)
    {
      SQLFreeEnv(cnc->henv);
      SQLFreeConnect(cnc->hdbc);
    }
  return 0;
}


Gda_ODBC_Recordset*
gda_odbc_connection_open_schema(Gda_ODBC_Connection*      cnc,
				Gda_ODBC_Error*           e,
				GDA_Connection_QType t,
				GDA_Connection_Constraint*      constraints,
				gint                 length)
{
  Gda_ODBC_Recordset* rc = gda_odbc_recset_new();
  schema_ops_fn       fn = schema_ops[t];

  cnc->recsets = g_list_append(cnc->recsets, rc);
  
  if (fn)
    {
      return fn(rc, e, cnc, constraints, length);
    }
  else
    {
      g_log("GDA ODBC", G_LOG_LEVEL_INFO,"schema_open_schema: Unhandled SCHEMA_QTYPE %d\n", t);
    }
  g_list_remove(cnc->recsets, rc);
  gda_odbc_recset_free(rc);
  return 0;
}



gint
gda_odbc_connection_close(Gda_ODBC_Connection* cnc)
{
  while (cnc->recsets)
    {
      gda_odbc_recset_close(cnc->recsets->data);
      cnc->recsets = g_list_remove_link(cnc->recsets, cnc->recsets);
    }
  while (cnc->commands)
    {
      gda_odbc_cmd_set_connection(cnc->commands->data, 0);
      cnc->commands = g_list_remove_link(cnc->commands, cnc->commands);
    }
  while(cnc->errors)
    {
      gda_odbc_error_free(cnc->errors->data);
      cnc->errors = g_list_remove_link(cnc->errors, cnc->errors);
    }
  SQLDisconnect(cnc->hdbc);
#if 0
  SQLFreeConnect(cnc->hdbc);
  SQLFreeEnv(cnc->henv);
#endif
  cnc->hdbc = 0;
  cnc->henv = 0;
  return 0;
}

